/*
 * misc.c --
 *
 * $Header: misc.c,v 1.60 88/07/15 10:30:37 t33876w Exp $
 *
 * Author: Jyrki Kuoppala (jkp@hutcs.hut.fi)
 *
 * Copyright 1988 by Helsinki University of Technology, Nixula
 *
 * Created: Wed Apr 13 09:26:48 1988
 * Last modified: Thu Nov  7 22:13:58 1991 jtp
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/types.h>
#ifdef BSD
# include <strings.h>
#else
# include <string.h>
#endif
#include <time.h>
#include "rmsg.h"
#include "rmsgprot.h"

char *append(), *ctime(), *malloc(), *strcpy();
long time();

char *basename(x)
char *x;
{
     char *tmp;
     if (tmp = rindex (x, '/')) tmp++;
     else tmp = x;
     return (tmp);
}

/*
 * This routine takes the header line and parses it to format in headerline
 * It returns a string, which should then be freed by the caller
 * If there is no free memory, this returns 0
 * 
 * The parameter 'secure' is used with the command 'exec'
 * when the result is given to 'sh -c'
 * If secure is set, every %? fromat char will be quoted
 * with single quotes for security and single quotes in client's
 * name or gecos field will be changed to '*'
 *
 * %f : the sender of the message in the form user@host
 * %n : the real name of the sender
 * %o : office
 * %p : phone
 * %h : home phone
 * %d : date when the message arrives
 */

char *parseheader (headerline, from, fromname, option)
char *headerline, *from, *fromname;
int option;
{
     char *buf, *tmp, *p;
     /* 200 should be enough for the gecos field */
     static char x[200];
     long now;
     int g;
     char *gecos[4];        /* for gecos fields */

     for (g = 0; g < 4; g++)
          gecos[g] = 0;
     g = 0;
     (void) strcpy (x, fromname);
     /* Real name */
     gecos[g++] = x;
     tmp = x;
     while (*tmp) {
          if (*tmp == ',') {
               *tmp = 0;
               if (g < 4)
                    gecos[g++] = tmp + 1;
          }
          tmp++;
     }
     /* OK, now we have gecos[] table containing gecos fields */

     /* Should be enough */
     if ((buf = malloc ((unsigned) strlen(headerline) + FROMNAMELEN +
                        FROMLEN + 500)) == NULL)
          return (0);
     tmp = buf;
     while (*headerline) {
          if (*headerline == '%') {
               switch (headerline++, *headerline++) {
               case 'f':                /* Append user@machine */
                    tmp = append (tmp, from, option);
                    break;
               case 'n':
                    tmp = append (tmp, gecos[0], option);
                    break;
               case 'o':
                    tmp = append (tmp, gecos[1], option);
                    break;
               case 'p':
                    tmp = append (tmp, gecos[2], option);
                    break;
               case 'h':
                    tmp = append (tmp, gecos[3], option);
                    break;
               case 'd':
                    now = time((long *) NULL);
                    p = ctime (&now);
                    /* We don't want the cr */
                    *(rindex (p, '\n')) = 0;
                    tmp = append (tmp, p, option);
                    break;
               default:
                    break;
               }
          }          
          else {
               if (option == ADD_CR && *headerline == '\n')
                    *tmp++ = '\r';
               *tmp++ = *headerline++;
          }
     }
     *tmp = 0;
     return (buf);
}

/*
 * Append string 'from' to string 'to', return the new 'to'
 * If option == SECURE, convert all single quotes to *:s
 * if option == ADD_CR, add \r:s before all \n:s
 */

char *append(to, from, option)
char *to, *from;
int option;
{
     if (option == SECURE) *to++ = '\047';
     if (from) {
          while (*from) {
               if (option == ADD_CR && *from == '\n')
                    *to++ = '\r';
               *to = *from++;
               if (option == SECURE && *to == '\047')
                    *to = '*';
               to++;
          }
     }
     if (option == SECURE) *to++ = '\047';
     return (to);
}               
