/*
 * parser.c -- read and parse the .msgconf file
 *
 * $Header: parser.c,v 1.61 88/08/20 19:55:42 t33876w Exp $
 *
 * Author: Jyrki Kuoppala (jkp@hutcs.hut.fi)
 *
 * Copyright 1988 by Helsinki University of Technology, Nixula
 *
 * Created: Mon Apr 11 01:51:03 1988
 * Last modified: Sat Aug 20 18:47:54 1988
 */

#include <stdio.h>
#if defined(SVR4)
# include <string.h>
#else
# include <strings.h>
#endif /* SVR4 */
#include <pwd.h>
#include <ctype.h>
#include <sys/types.h>
#include "rmsg.h"

/*
 * This part contains all the global configuration parameters set in 
 * read_config()
 */

struct ali *aliases;          /* Contains linked list of aliases */
struct ali *cur_ali;          /* Used during alias reading */

struct on_ttys *on_ttys;
struct on_ttys *cur_tty;

int denymsgs = 0;             /* Do we allow remote messages, default = yes */
int inappend = 0;             /* Append to incoming message file */
int single = 0;               /* If <> 0, write to only one tty */
#ifdef BITNET
char bitnetserver[LINELEN] = ""; /* name of the bitnet server user */
#endif
char inlast[PATHLEN]="";      /* File to store last incoming message,
                                 default: do not store */
char outlast[PATHLEN]=".msgout"; /* Last outgoing message */

char outheader[LINELEN]="";
char outlog[PATHLEN]="";         /* Log for outgoing messages */
char home[PATHLEN];
char headerline[LINELEN] = "\007\nMessage from %n <%f>:\n";
                              /* Line shown with incoming message */
char execline[LINELEN] = "";
char awayline[LINELEN] = "";

struct command comtab[] = {
     "alias",        f_alias,
     "away",         f_away,
#ifdef BITNET
     "bitnetserver", f_bitnetserver,
#endif
     "deny",         f_deny,
     "exec",         f_exec,
     "header",       f_header,
     "inlast",       f_inlast,
     "inappend",     f_inappend,
     "outheader",    f_outheader,
     "outlast",      f_outlast,
     "outlog",       f_outlog,
     "single",       f_single,
     NULL,           NULL
     };

/* end of config params */

struct passwd *getpwnam();
char *malloc();

extern char *progname;
extern int debug;

void freeargs();
int exec_func(), parse_line(), setflag(), setstring();

char home[1024];
extern int root;

int errors;

int read_config(user)
char *user;
{
     struct passwd *pwd;
     static char line[LINE_LEN];
     static char config[PATHLEN];

     static char *argv[MAX_ARGV];
     static int argc;
     struct ali *a;

     FILE *in;

     errors = 0;

     /* Allocate a dummy first element from which we start searches */

     if ((a = (struct ali *) malloc (sizeof (struct ali))) == NULL)
          exit (-1);
     a->next = 0;
     cur_ali = aliases = a;

     if (!(pwd = getpwnam(user)))
          return(ERR_UNK_USER);

     if (root) {
          (void) setgid (pwd->pw_gid);
          if (setuid (pwd->pw_uid) == -1)
	       (void) setuid(NOBODY);
     }

     (void) strcpy (home, pwd->pw_dir);
     (void) sprintf (config, "%s/%s", home, CONFIGFILE);
     
     if ((in = fopen (config, "r")) == NULL)
          return(0);

     while (fgets (line, LINE_LEN-1, in) != NULL) {
          (void) parse_line (line, &argc, &argv[0]);

          exec_func (argc, argv);
          freeargs (argc, argv);
     }
     (void) fclose (in);
     return(errors);
}

int f_outlast(argc, argv)
int argc;
char **argv;
{
     return (setstring (outlast, PATHLEN - 1, argc, argv));
}
     
int f_exec(argc, argv)
int argc;
char **argv;
{
     return (setstring (execline, LINELEN - 1, argc, argv));
}

int f_away(argc, argv)
int argc;
char **argv;
{
     return (setstring (awayline, LINELEN - 1, argc, argv));
}

int f_header(argc,argv)
int argc;
char **argv;
{
     return (setstring (headerline, LINELEN - 1, argc, argv));
}

int f_inlast(argc, argv)
int argc;
char **argv;
{
     return (setstring (inlast, PATHLEN - 1, argc, argv));
}

int f_outlog(argc, argv)
int argc;
char **argv;
{
     return (setstring (outlog, PATHLEN - 1, argc, argv));
}

int f_outheader(argc, argv)
int argc;
char **argv;
{
     return (setstring (outheader, PATHLEN - 1, argc, argv));
}

#ifdef BITNET
int f_bitnetserver(argc, argv)
int argc;
char **argv;
{
     return (setstring (bitnetserver, PATHLEN - 1, argc, argv));
}
#endif

int f_deny(argc, argv)
int argc;
char **argv;
{
     return (setflag (&denymsgs, argc, argv));
}

int f_inappend(argc, argv)
int argc;
char **argv;
{
     return (setflag (&inappend, argc, argv));
}

int f_single(argc, argv)
int argc;
char **argv;
{
     return (setflag (&single, argc, argv));
}

/* Called when we encounter a 'alias' command in .msgconf */

int f_alias(argc,argv)
int argc;
char **argv;
{
     struct ali *a;
     char *map,*to;
     if (debug) {
          (void) fprintf(stderr,"%s: called alias, argc: %d\n", progname,
                         argc);
     }
     if (argc != 3) return (-1);
     if ((a = (struct ali *) malloc (sizeof (struct ali))) == NULL)
          return (-1);
     if ((map = malloc ((unsigned) strlen (argv[1]) + 1)) == NULL )
          return (-1);
     if ((to = malloc ((unsigned) strlen (argv[2]) + 1)) == NULL )
          return (-1);
     (void) strcpy (map, argv[1]);
     (void) strcpy (to, argv[2]);
     a->map = map;
     a->to = to;
     a->next = 0;
     cur_ali = cur_ali->next = a;
     return (0);
}

/*
 * Set a string according to argc and argv
 * If no arguments, set string empty
 */

int setstring (string, len, argc, argv)
char *string;
int argc, len;
char **argv;
{
     if (argc == 1)
          *string = 0;
     else {
          if (argc > 2)
               return(-1);
          (void) strncpy (string, argv[1], len);
     }
     return (0);
}

/*
 * Set a flag according to argc and argv
 */

/* ARGSUSED */
int setflag(flag, argc, argv)
int *flag;
int argc;
char **argv;
{
     if (argc != 1)
          return (-1);
     else *flag = 1;
     return (0);
}

/*
 * Called to get a match from the alias list
 * If found, returns a pointer to it ; if not, returns it's argument
 */

char *getalias(receiver)
char *receiver;
{
     struct ali *a = aliases;
     while (a = a->next) {
          if (strcmp (a->map, receiver) == 0)
               return (a->to);
     }
     return (receiver);
}

/*
 * Take a line as input and parse it to argc and **argv
 * Parse |t, |020 and "a b" 
 *
 * Quite an ugly routine, don't look at it .. :-(
 */

int parse_line(line, argc, argv)
char *line;
int *argc;
char **argv;
{
     char *p;
     static char tmp[LINE_LEN];
     int quote=0, lastquote=0, oct=0;

     *argc = 0;

     p = tmp;

     while (*line == ' ' || *line == '\t') line++;
     
     while (*line) {
          switch (*line) {
          case '\\':
               lastquote = 0;
               switch (*++line) {
               case '\\':
                    *p++ = '\\';
                    break;
               case 'r':
                    *p++ = '\r';
                    break;
               case 't':
                    *p++ = '\t';
                    break;
               case 'n':
                    *p++ = '\n';
                    break;
               case 'b':
                    *p++ = '\b';
                    break;
               default:
                    if (isdigit(*line)) {
                         oct = ((*line - '0') << 6);
                         oct += ((*++line - '0') << 3);
                         oct += (*++line - '0');
                         *p++ = oct;
                    }
                    else
                         *p++ = *line;
               }
               break;
          case '\047':                  /* single quote */
          case '\042':                  /* double quote */
               quote = 1 - quote;
               if (lastquote) {
                    *p++ = *line;
                    lastquote = 0;
               }
               else lastquote = 1;
               break;
          case ' ':
          case '\t':
          case '\n':
          case '#':
               lastquote = 0;
               if (quote && (*line != '\n')) {
                    *p++ = *line;
               }
               else {
                    /* End of this argument */
                    *p = 0;
                    p = malloc ((unsigned) strlen(tmp)+1);
                    (void) strcpy (p, tmp);
                    argv[(*argc)++] = p;
                    p = tmp;
                    if (*line == '#' || *line == '\n') *line = 0; /* eol */

                    /* Find start of next word */
                    else {
                         line++;
                         while (*line == ' ' || *line == '\t') line++;
                    }
                    if (*line == '#' || *line == '\n') *line = 0; /* eol */
                    line--;
               }
               break;
          default:
               lastquote = 0;
               *p++ = *line;
               break;
          }
          line++;
     }
     argv[*argc] = 0;
     return(0);
}

/* Free the malloc()ed arguments */

void freeargs (argc, argv)
int argc;
char **argv;
{
     int i;
     for (i = 0; i < argc ; i++) free(argv[i]);
}

/*
 * Execute conf file function when passed arguments as to main()
 */

int exec_func(argc,argv)
int argc;
char **argv;
{
     struct command *i;

     for (i=comtab; i->token != NULL ; i++) {
          if (debug) {
               (void) fprintf (stderr,"%s\n",i->token);
          }
          if (strcmp (argv[0], i->token) == 0) {
               if ((i->function)(argc,argv) == -1) {
                    errors++;
               }
          }
     }
}
