/*
 * rmsg.c -- send a message via RPC to a remote machine
 *
 * $Header: rmsg.c,v 1.63 88/08/20 18:11:13 t33876w Exp $
 * $Log:	rmsg.c,v $
 * Revision 1.63  88/08/20  18:11:13  t33876w
 * Fixed a bug with SysV uid & gid sizes
 * also now unknown error is correctly recognized
 * 
 * Revision 1.62  88/07/22  08:25:47  t33876w
 * Added dollar Log 
 * 
 * 
 *
 * Author: Jyrki Kuoppala (jkp@hutcs.hut.fi)
 *
 * Copyright 1988 by Helsinki University of Technology, Nixula
 * You may use this program in the GNU spirit
 *
 * Thanks to Sun, on whose tutorial this is based on.
 *
 * Created: Sun Apr 10 17:36:07 1988
 * Last modified: Sat Aug 20 16:53:00 1988
 */
#include <stdio.h>
#include <rpc/rpc.h>
#ifdef BSD
#include <strings.h>
#else
#include <string.h>
#endif

#include <pwd.h>
#include <sys/file.h>
#ifdef USG5
# include <fcntl.h>
#endif
#include <netdb.h>
#include "rmsg.h"
#include "rmsgprot.h"

#define HNLEN 200

int xdr_rmsg(), getopt();
char *getlogin(), *basename(), *index(), *rindex();
uid_t getuid();
gid_t getgid();
struct passwd *getpwuid(), *getpwnam();
void report();
char *getalias(), *parseheader();
#ifdef SUNOS
void *malloc();
#else
char *malloc();
#endif
void storeout();
extern struct hostent *gethostbyname();

extern char outlast[], home[], outheader[], outlog[];
#ifdef BITNET
extern char bitnetserver[];
#endif

char *progname;
int debug = 0;
int root = 0;

char a[MSGLEN+1];

#ifndef lint
static char rcsid[] = "$Header: rmsg.c,v 1.63 88/08/20 18:11:13 t33876w Exp $";
#endif

main(argc,argv)
int argc;
char **argv;
{
     extern char *optargs;
     extern int optind;

     static char file[PATHLEN];

     int len,c=0, resend = 0, errflag = 0, exitval = 0;
     unsigned int status;
     char *place=a, *miu, *tmp, *tmp2, hostn[HNLEN], *host, *user, *s;
     char from[FROMLEN+1];
#ifdef BITNET
     char bitdest[FROMNAMELEN+1];
#endif
     struct passwd *pwd;
     char fromname[FROMNAMELEN+1];

     struct rmsg_par rmsgargs;
     struct hostent *hp = NULL;

     progname = argv[0];

#ifdef lint
     /* fromname is used later in strcpy, this is to satisfy lint :-) */
     *fromname = 0;
#endif
     while ((c = getopt(argc, argv, "dr")) != EOF) {
          switch (c) {
          case 'd':
               debug++;
               break;
          case 'r':
               resend++;
               break;
          case '?':
          default:
               errflag++;
               break;
          }
     }

     if (errflag || optind >= argc) {
          (void) fprintf(stderr,"usage: %s [-r] [-d] user[@host][,user[@host]] ...\n",
                         progname);
          exit(1);
     }
     (void) gethostname(hostn, HNLEN-1);
     if (hp = gethostbyname(hostn)) {
       strncpy(hostn,hp->h_name,HNLEN-1);
     }

     if ((tmp = getlogin()) == NULL) {
          pwd = getpwuid((int) getuid());
          tmp = (pwd) ? pwd->pw_name : "intruder";
     }
     user = malloc ((unsigned int) strlen (tmp) + 1);
     (void) strcpy (user, tmp);
     pwd = getpwnam (user); /* get name */
     
     if (pwd != 0) {
          tmp = pwd->pw_gecos;
     }
     else tmp = "Ike Intruder";

     /* Store real name for later usage */
     (void) strcpy (fromname, tmp);

     (void) sprintf(from, "%s@%s", user, hostn);
     if (read_config (user) == -1) {
          (void) fprintf(stderr, "%s: warning: errors reading config file\n",
                  progname);
     }
     if (resend) {
          (void) sprintf (file, "%s/%s", home, outlast);
          if ((resend = open (file, O_RDONLY)) < 0) {
               (void) fprintf (stderr, "%s: Can't open %s\n",
                               progname, outlast);
               exit (-1);
          }
     }
     /* Resend is 0 or the msglast file */
     do {
          len=read(resend, place, MSGLEN-(int)(place-a));
          place += len;
     } while ((len > 0) && (place < a+MSGLEN));

     if (fork()) _exit(0);  /* Deliver in background, since it might take
                             a while */

     for (; optind < argc; optind++) {
          char *s2 = index(argv[optind],',');
	  if (s2) {	/* Split at commas, reiterate.. */
	    char *s1 = argv[optind];
	    argv[optind] = s2+1;
	    argv[optind-1] = s1;
	    *s2 = 0;
	    optind -= 2;
	    continue;
	  }
	  tmp = getalias (argv[optind]);
          storeout (a, tmp);
#ifdef BITNET
          /*
           * If bitnet is used, rmsgargs.to contains the destination
           * bitnet address without the .bitnet postfix, ie. jkp@finhutcs
           * Rmsg daemon which runs on the bitnet server host
           * sees the @ in the name of the receiver and forwards
           * the message to bitnet.
           * Host contains the string which is set by the bitnetserver
           * option of .msgconf.
           */

          if ((*bitnetserver) && 
              (strcmp (BITNETPOSTFIX, (tmp2 = rindex (tmp, '.')) + 1)) == 0) {
               *tmp2++ = '\0';
               rmsgargs.to = tmp;
               host = bitnetserver;
          }
          else {
#endif
               if ((miu = rindex(tmp, '@')) == 0) {
                    /* Use our hostname, if not specified */
                    host = hostn;
               }
               else {
                    *miu = '\0';
                    host = miu + 1;
               }
               rmsgargs.to = tmp;
#ifdef BITNET
          }
#endif

          rmsgargs.from = from;
          rmsgargs.fromname = fromname;
          rmsgargs.msg = a;

          status = 0;
          if (callrpc (host, RMSGPROG, RMSGVERS, RMSGPROC_MSG,
                       xdr_rmsg, &rmsgargs, xdr_u_int, &status) != 0) {
               (void) fprintf(stderr, "%s: error\007: can't connect to server\n",
                              progname);
               exitval++;
          }
          report (status, &rmsgargs);
     }
     exit(exitval);
}

/* Report if the rpc call succeeded or not */
void report (status, msg)
unsigned int status;
struct rmsg_par *msg;
{
     switch (status) {
     case 0:
          break;
     case ERR_REFUSING:
          (void) fprintf(stderr, "%s: \007%s refusing messages\n",progname,
                         msg->to);
          break;
     case ERR_UNK_USER:
          (void) fprintf(stderr, "%s: \007unknown user %s\n", progname,
                         msg->to);
          break;
     case ERR_NOTON:
          (void) fprintf(stderr, "%s: \007%s not logged on\n",
                         progname, msg->to);
          break;
     case -1:
     case ERR_UNKNOWN:
     default:
          (void) fprintf(stderr, "%s: \007unknown remote error\n", progname);
          break;
     }
}

void storeout (msg, to)
char *msg, *to;
{
     static char outfile[PATHLEN];
     char *buf;
     int storefile;
     
     if (*outlast) {
          (void) sprintf (outfile, "%s/%s", home, outlast);
          if ((storefile = open (outfile, O_WRONLY | O_CREAT
                                 | O_TRUNC, 0600)) > 0) {
               (void) write (storefile, msg, strlen(msg));
               (void) close (storefile);
          }
     }
     if (*outlog) {
          (void) sprintf (outfile, "%s/%s", home, outlog);
          if ((storefile = open (outfile, O_WRONLY | O_CREAT
                                 | O_APPEND, 0600)) > 0) {
               if (buf = parseheader (outheader, to, "", 0)) {
                    (void) write (storefile, buf, strlen(buf));
                    (void) free (buf);
               }
               (void) write (storefile, msg, strlen(msg));
               (void) close (storefile);
          }
     }
}
