/*
 * server.c -- server for the remote message system using rpc
 *
 * $Header: server.c,v 1.63 88/08/20 18:12:18 t33876w Exp $
 * $Log:	server.c,v $
 * Revision 1.63  88/08/20  18:12:18  t33876w
 * Fixed a bug with SysV uid & gid sizes
 * also now unknown error is correctly recognized
 * 
 * Revision 1.62  88/07/22  08:27:51  t33876w
 * No longer takes basename of 'inlast' file.
 * 
 *
 * Author: Jyrki Kuoppala (jkp@hutcs.hut.fi)
 *
 * Copyright 1988 by Helsinki University of Technology, Nixula
 * You may use this program in the GNU spirit
 *
 * Thanks to Sun, on whose tutorial this is based on.
 *
 * Created: Sun Apr 10 16:11:02 1988
 * Last modified: Fri Jul 22 00:50:49 1988
 */

#define SERVER

#include <stdio.h>
#if defined(SVR4)
# include <netinet/in.h>
#endif /* SVR4 */
#include <rpc/rpc.h>
#if defined(SVR4)
# include <rpc/svc_soc.h>
#endif /* SVR4 */
#include "rmsgprot.h"
#include "rmsg.h"

#include <utmp.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/file.h>
#if defined(SVR4)
# include <string.h>
#else
# include <strings.h>
#endif /* SVR4 */
#include <sys/ioctl.h>
#ifndef USG5
#include <sys/wait.h>
#endif

#if defined(USG5) || defined(SVR4)
#include <fcntl.h>
#endif /* USG5 || SVR4 */

#include <sys/param.h>
#include <sgtty.h>
#include <sys/types.h>
#include <ctype.h>

int xdr_rmsg(), smsg();
int tty_write(), dowrite();
char *basename(), *parseheader();
#ifdef SUNOS
void *malloc();
#else
char *malloc();
#endif
long time();
uid_t getuid();
char *progname;
int debug = 0;

extern int denymsgs, inappend, single;
extern char headerline[], inlast[], home[], execline[], awayline[];

extern struct on_ttys *on_ttys;
extern struct on_ttys *cur_tty;

#ifndef lint
static char rcsid[] = "$Header: server.c,v 1.63 88/08/20 18:12:18 t33876w Exp $";
#endif

int root; /* True if server started as root */

main(argc,argv)
int argc;
char **argv;
{
     SVCXPRT *transp;
     int fd;

     progname=argv[0];

     if (getuid () == 0) root++;

     if (! root) {
          (void) fprintf (stderr, "%s: not started as root, some features not available\n", progname);
     }

     if (! strcmp (basename (progname), "rmsgd")) {
          if (fork()) _exit(0);  /* make myself a daemon */
#ifdef BSD
          fd = open("/dev/tty", O_RDWR);  /* disconnect from control tty */
          if (fd >= 0) {
               (void) ioctl(fd, TIOCNOTTY, (char *)0);
               (void) close(fd);	/* close this, don't need any more*/
          }
#else
          (void) setpgrp ();
#endif
     }

     else {
          (void) fprintf (stderr, "Starting from inetd not yet supported, rename me ase rmsgd and start\nfrom /etc/rc\n");
     }

     if (argc > 1)
          if (strcmp(argv[1], "-d") == 0) {
               debug++;
               argv++;
          }

     transp = svcudp_create(RPC_ANYSOCK);
     if (transp == NULL) {
          (void) fprintf(stderr,"%s: can't create an RPC server\n",progname);
          exit(-1);
     }
     pmap_unset(RMSGPROG, RMSGVERS);
     if (!svc_register (transp, RMSGPROG, RMSGVERS, smsg, IPPROTO_UDP)) {
          (void) fprintf(stderr,"%s: can't register RMSG service\n",progname);
          exit(-2);
     }
     svc_run();
     (void) fprintf(stderr,"%s: svc_run returned !\n",progname);
     exit(-3);
}

smsg(rqstp, transp)
struct svc_req *rqstp;
SVCXPRT *transp;
{
     int child, pid, fd;
     unsigned int retval = 0;
#if defined(HPUX) || defined(SVR4)
     int status;
#else
     union wait status;
#endif
     struct rmsg_par rmsgargs;

     if (debug) 
          printf("%s: incoming RPC call, proc: %d\n",progname,rqstp->rq_proc);

     /* rqstp->rq_proc is unsigned long */
     switch ((int) rqstp->rq_proc) {
     case NULLPROC:
          if(!svc_sendreply(transp, xdr_void, (char *) 0)) {
               (void) fprintf(stderr,"%s: can't reply to RPC call\n",progname);
               exit(1);
          }
          return;
     case RMSGPROC_MSG:
          /* OK, this is time to start working :-) */
          /* First zero the args so xdr_rmsg alloc's them for us */
          (void) bzero ((char *) &rmsgargs, sizeof (rmsgargs));
          if (!svc_getargs(transp, xdr_rmsg, &rmsgargs)) {
               svcerr_decode(transp);
               return;
          }
          
          /* We need to fork because otherwise we would get a controlling
             tty */

          if ((child = fork()) == 0) {
#ifdef BSD
               (void) signal(SIGTTOU,SIG_IGN);
#endif
               if (! debug) {
                    (void) close (0);
                    (void) close (1);
                    (void) close (2);
               }
                                /* we should't have std input or output */
                                /* now */
               fd = open ("/dev/null", O_RDWR);
               (void) dup (fd);
               (void) dup (fd);
               _exit (dowrite (rmsgargs.to, rmsgargs.from, rmsgargs.fromname,
                              rmsgargs.msg));
          }
          pid = wait (&status);
          if (pid != child) {
               (void) fprintf (stderr, "%s: strange, wait returned %d instead of %d\n",
                        progname, pid, child);
               exit(-1);
          }
#if defined(HPUX) || defined(SVR4)
          if (status & 0xff) retval = -1;
          else retval = ((status & 0xffff) >> 8);
#else
          if (status.w_termsig) retval = -1;       /* ended in signal */
          else retval = status.w_retcode;
#endif /* HPUX || SVR4 */
          if (debug) (void) fprintf (stderr,"retval: %u\n", retval);

          if (!svc_freeargs(transp, xdr_rmsg, &rmsgargs)) {
               svcerr_decode(transp);
               return;
          }

          if (!svc_sendreply(transp, xdr_u_int, (char *) &retval)) {
               (void) fprintf(stderr, "%s: can't reply to RPC call\n",
                              progname);
               exit(1);
          }
     }
}

/* 
 * Takes as parametres the following :
 * from whom the message is, to whom it is and the msg itself.
 * Returns 0 on success, ERR_NOTON if receiver is not logged on, ERR_REFUSING
 * if receiver is refusing messages and ERR_UNKNOWN on other errors
 * It calls tty_write on every line it finds the user is logged on.
 */

int dowrite(to, from, fromname, msg)
char *from, *fromname, *to, *msg;
{
     struct stat statbuf;
     char tty[UTMP_TTY_SIZE+6];
     int ut, status = ERR_NOTON; /* ttycount = 0; */
     struct utmp utent;
     struct on_ttys *a;

     struct on_ttys *single_tty = (struct on_ttys *) 0;
     time_t touch_time = (time_t) 0;

     FILE *com;
     char lastfile[PATHLEN], *buf;
     int storefile;

#ifdef BITNET
     char nwline[60];
     /* First check if the message is addressed to bitnet.  If it */
     /* is, forward it to the urep daemons */
     
     if (index (to, '@')) {
          (void) sprintf (nwline, "echo netwrite %s >>/tmp/netwrite", to);
          system (nwline);
          return (0);
     }
#endif

     /* On server side we don't care about config errors other than */
     /* to announce unknown user */
     if (read_config (to) == ERR_UNK_USER)
          return (ERR_UNK_USER);

     if (root) {
          /* Don't do any error checking on writing to file, since it's
             not reasonable to report it anyway */
          
          /* Handle pathnames relative to the home dir of the
             recipient */
          if (chdir (home) < 0) return (ERR_UNKNOWN);
          
          /* Write to 'lastfile' if needed */
          if (*inlast) {
               (void) sprintf (lastfile, "%s/%s", home, inlast);
               if ((storefile = open (lastfile,
                                      inappend ? O_WRONLY | O_CREAT | O_APPEND :
                                      O_WRONLY | O_TRUNC | O_CREAT, 0600)) > 0) {
                    if (buf = parseheader (headerline, from, fromname, 0)) {
                         (void) write (storefile, buf, strlen(buf));
                         (void) free (buf);
                         (void) write (storefile, msg, strlen(msg));
                    }
                    (void) close (storefile);
               }
          }
          if (*execline) {
               status = 0;
               /* We have to fork for the exec, since we don't know how long
                  it takes */
               if (fork()) return(status);
               if (buf = parseheader (execline, from, fromname, SECURE)) {
                    if ((com = popen (buf, "w"))) {
                         (void) fprintf (com, "%s", msg);
                         (void) pclose (com);
                    }        
                    (void) free (buf);
               }
               /* If we exec a command when it comes, we don't write it
                  to tty */
               return(status);
          }
     }

     if ((ut = open(UTMP, O_RDONLY)) < 0) {
          (void) fprintf(stderr,"%s: couldn't open utmp file\n",progname);
          perror("open");
          return(ERR_UNKNOWN);
     }
     /* Allocate a dummy first element from which we start searces */

     if ((a = (struct on_ttys *) malloc (sizeof (struct on_ttys))) == NULL)
          _exit (-1);
     a->next = 0;
     cur_tty = on_ttys = a;

     while (read (ut, (char *) &utent, sizeof(utent)) ==
            sizeof (utent)) {
          if (strncmp (to, utent.ut_name, UTMP_NAME_SIZE) == 0) {
               /* utent.ut_line might not be null terminated */
#if defined(USG5) || defined(SVR4)
               if (utent.ut_type != USER_PROCESS)
                    continue;
#endif /* USG5 || SVR4 */
               (void) strcpy (tty, "/dev/");
               (void) strncat (tty, utent.ut_line, UTMP_TTY_SIZE);

               if (stat (tty, &statbuf) == -1)
                    continue;
               if (!(statbuf.st_mode & S_IFCHR))
                    continue;           /* A little paranoia .. */
               if (!(statbuf.st_mode & 020)) {
                    /* no write permission for group */
                    /* If we get thru on one tty, we return success */
                    if (status) status = ERR_REFUSING;
                    continue;
               }

               status = 0;
               /* Now we think we can write to the tty */
               
	       /* Check if we really have write access */
	       if ((! root) && ((gid_t) statbuf.st_gid != getgid()))
		    return (ERR_UNKNOWN);

               if ((a = (struct on_ttys *) malloc (sizeof (struct on_ttys)))
                   == NULL)
                    return (ERR_UNKNOWN);
               (void) strcpy (a->tty, tty);
     
               if ((touch_time == 0) || (touch_time < statbuf.st_atime)) {
                    touch_time = statbuf.st_atime;
                    single_tty = a;
               }

               a->next = 0;
               cur_tty = cur_tty->next = a;
/*               ttycount++;*/
          }
     }
     (void) close(ut);

     /* If recipient is logged on but denying messages by 'deny'
        we handle it as if he is not logged on */

     if (denymsgs && status == 0) {
          status = ERR_REFUSING;
          return (status);
     }
     if (root) {
          /* If he is not logged on, exec 'away' command if it is set */
          if (*awayline && status == ERR_NOTON) {
               /* We have to fork for the exec, since we don't know how long
                  it takes */
               if (fork()) return(status);
               if (buf = parseheader (awayline, from, fromname, SECURE)) {
                    if ((com = popen (buf, "w"))) {
                         (void) fprintf (com, "%s", msg);
                         (void) pclose (com);
                    }        
                    (void) free (buf);
               }
               /* If we exec a command when it comes, we don't write it
                  to tty */
               return(status);
          }
     }

     /* Fork for the actual writing */
     if (fork()) return (status); 
     
     a = on_ttys;
     if (! status)
          while (a = a->next) {
               if (a == single_tty || (! single))
                    (void) tty_write (a->tty, from, fromname, msg);
          }
     return (status);
}

/*
 * Takes the line with the leading /dev, from whom the message
 * is and the message itself.
 * Returns 0 on success
 * ERR_UNKNOWN on various errors
 */

int tty_write (line, from, fromname, msg)
char *line, *from, *fromname, *msg;
{
     int fp;
     char *out, *p;

     if ((fp = open (line, O_WRONLY)) < 0)
          return(ERR_UNKNOWN);
     /* Add carriage returns in front of line feeds */
     if ((out = malloc ((unsigned int) strlen (msg) * 2)) == 0)
          return (ERR_UNKNOWN);
     p = out;
     while (*msg) {
          if ((*msg & 0x7f) == '\n') {
               *p++ = '\r';
               *p++ = '\n';
          }
          else {
               if (isprint (*msg & 0x7f) || ((*msg & 0x7f) == '\t'))
                    *p++ = *msg & 0x7f;
               else
                    *p++ = '?';
          }
          msg++;
     }
     *p = 0;
     if (p = parseheader (headerline, from, fromname, ADD_CR)) {
          (void) write (fp, p, strlen(p));
          (void) free (p);
          if (write (fp, out, strlen(out)) < 0) {
               (void) close (fp);
               (void) free (out);
               return(ERR_UNKNOWN);
          }
     }
     else {
          (void) close (fp);
          (void) free (out);
          return (ERR_UNKNOWN);        /* malloc in parseaddr failed */
     }
     (void) free (out);
     (void) close (fp);
     return(0);
}
